//////////////////////////////////////////////////////////////////////
// Simple ARM9 demo (touch screen controls the colors)
// -- joat
//////////////////////////////////////////////////////////////////////
//
#include <stdio.h>

#include <NDS/NDS.h>

#include <NDS/ARM9/rand.h>
#define REG_VCOUNT *(vu16*)0x4000006		//Our good old vblank counter

#include <string.h>
#include <ctype.h>

#include "depack.h"

#include "collision.h"

#include "build/bird_raw.h"
#include "build/bullet_raw.h"
#include "build/man_raw.h"
#include "build/poop_raw.h"
#include "build/sprites_pal.h"
#include "build/power_hp_raw.h"
#include "build/power_spd_raw.h"
#include "build/power_shot_raw.h"

#include "build/font_raw.h"
#include "build/font_pal.h"

#define abs(x) (((x)<0)?(-(x)):((x)))

#define SPR_BULLET 0
#define SPR_BIRD 2
#define SPR_MAN 18
#define SPR_POOP 34                            
#define SPR_POWER_HP 38
#define SPR_POWER_SHOT 46
#define SPR_POWER_SPD 54

                                 

#define DEB                            

//////////////////////////////////////////////////////////////////////
//-------------------------------------------------------------------
void
print(char *s)
{
#ifdef DEB

    // warning! will crash on hardware!
    asm volatile ("mov r0, %0;" "swi 0xff0000;":        
                  :"r" (s):"r0");

#endif    
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//
//

#define TOUCH_X (IPC->touchX / 14)
#define TOUCH_Y (IPC->touchY / 18 - 12)

#define TOUCH_2X (IPC->touchX / 7)
#define TOUCH_2Y (IPC->touchY / 9 - 24)

#define KEY_TOUCH (((~IPC->buttons) << 6) & (1<<12))

inline void COPY16(uint16* from, uint16* to,int size)
{
    uint8 *from8 = (uint8*)from;
    uint8 *to8 = (uint8*)to;
    for(int i=0;i<size*2;i++)
        to8[i]=from8[i];
}

void writeChar(int x, int y, char c, uint16 *map)
{
    map[y*32 + x] = c - ' ';
}

void writeNum(int x, int y, int num, uint16 *map)
{
    if(num < 100)
        map[y*32 + x] = 0;
    else
        map[y*32 + x] = '0' - ' ' + ((num/100) % 10);

    if(num < 10)
        map[y*32 + x+1] = 0;
    else
        map[y*32 + x+1] = '0' - ' ' + ((num/10) % 10);

    map[y*32 + x+2] = '0' - ' ' + (num % 10);
}

void writeText(int x, int y, char *str, uint16 *map)
{
    for(int xp=0; xp<strlen(str); xp++)
    {
        int xx = x+xp;
        int yy = y;
        while(xx >= 32)
        {
            xx -= 32;
            yy++;
        }
        writeChar(xx, yy, toupper(str[xp]), map);
    }
}


///////////////////////

//OAM globals
sSpriteEntry OAMCopy[128];
sSpriteEntry OAMCopySub[128];

//Sprite structure
typedef struct
{
	int x,y;	//location 
	int xSpeed, ySpeed;	//speed
        bool alive;
        int lastbomb;
	int oam;
        int hp;
        uint16 attribute[3];

}Sprite;

int score=0;
int hits=0;
int fired=0;
int manspeed = 1;
int mangun = 1;

//-------------------------------------------------------------------

//Moves a sprite in OAM memory
void MoveSprite(Sprite* sp)
{
	int x = sp->x >> 8;
	int y = sp->y >> 8;
        sSpriteEntry *oam;

        // only draw on top screen if y < 192
        oam = &OAMCopySub[sp->oam];
        if(sp->alive && y <= 192)
        {
            oam->attribute[0] = sp->attribute[0];
            oam->attribute[1] = sp->attribute[1];
            oam->attribute[2] = sp->attribute[2];

            oam->attribute[1] &= 0xFE00;
            oam->attribute[1] |= (x & 0x01FF);

            oam->attribute[0] &= 0xFF00;
            oam->attribute[0] |= (y & 0x00FF);
        }
        else
            oam->attribute[0] = ATTR0_DISABLED;

        // draw on bottom screen if y is > 128 (that is 192 - 64, the tallest possible sprite)
        oam = &OAMCopy[sp->oam];
        if(sp->alive && y >= 128)
        {
            y -= 192;

            oam->attribute[0] = sp->attribute[0];
            oam->attribute[1] = sp->attribute[1];
            oam->attribute[2] = sp->attribute[2];

            oam->attribute[1] &= 0xFE00;
            oam->attribute[1] |= (x & 0x01FF);

            oam->attribute[0] &= 0xFF00;
            oam->attribute[0] |= (y & 0x00FF);
        }
        else
            oam->attribute[0] = ATTR0_DISABLED;
} 

//-------------------------------------------------------------------
//Copies OAM buffer to actual OAM memory space
void updateOAM(void)
{
	for(int i = 0; i < 128 * sizeof(sSpriteEntry) / 4 ; i++)
	{
		((uint32*)OAM)[i] = ((uint32*)OAMCopy)[i];
		((uint32*)OAM_SUB)[i] = ((uint32*)OAMCopySub)[i];
	}
}

//Initialises OAM, moving all sprites offscreen
void initOAM(void)
{
	for(int i = 0; i < 128; i++)
	{
		OAMCopy[i].attribute[0] = ATTR0_DISABLED;
		OAMCopySub[i].attribute[0] = ATTR0_DISABLED;
		OAMCopy[i].attribute[1] = 0;
		OAMCopySub[i].attribute[1] = 0;
		OAMCopy[i].attribute[2] = 0;
		OAMCopySub[i].attribute[2] = 0;
	}	
        updateOAM();
}

//-------------------------------------------------------------------



	Sprite birds[40];
	Sprite bombs[40];
        Sprite shots[40];
        Sprite pups[5];
        Sprite man;

void dropBomb(Sprite *bird)
{
        int n=-1;
    for(int i=0; i<40; i++)
    {
        if(bombs[i].alive == false)
        {
            n=i;
            break;
        }
    }
    if(n==-1)
        return;

    bombs[n].alive = true;

    bombs[n].attribute[0] = ATTR0_COLOR_256 | ATTR0_TALL;
    bombs[n].attribute[1] = ATTR1_SIZE_8;
    bombs[n].attribute[2] = SPR_POOP;

    bombs[n].x = bird->x + (16<<8);
    bombs[n].y = bird->y;

    bombs[n].xSpeed = bombs[n].ySpeed = 0;

    bombs[n].oam = n + 80;

    bird->lastbomb = 0;
    

    MoveSprite(&bombs[n]);
}

void moveBombs()
{
    for(int bm=0; bm<40; bm++)
    {
        if(bombs[bm].alive)
        {
            bombs[bm].ySpeed += 1<<5;
            bombs[bm].y += bombs[bm].ySpeed;

            if(bombs[bm].y > (192*2)<<8)
                bombs[bm].alive = false;

            MoveSprite(&bombs[bm]);
        }
    }
}
        



void addBird()
{
    int n=-1;
    for(int i=0; i<40; i++)
    {
        if(birds[i].alive == false)
        {
            n=i;
            break;
        }
    }
    if(n==-1)
        return;

    if(rand() % 2 == 0)
    {
        birds[n].x = (-33) << 8;
        birds[n].xSpeed = (rand() % (1<<8))+(1<<8);
    }
    else
    {
        birds[n].x = (256) << 8;
        birds[n].xSpeed = -((rand() % (1<<8))+(1<<8));
    }

    birds[n].alive = true;

    birds[n].y = (rand() % (192<<8));
    birds[n].ySpeed = 0;
    birds[n].oam = n;

    birds[n].lastbomb = 0;

    birds[n].attribute[0] = ATTR0_COLOR_256 | ATTR0_WIDE;
    birds[n].attribute[1] = ATTR1_SIZE_32;
    birds[n].attribute[2] = SPR_BIRD;
    

    MoveSprite(&birds[n]);
}


void moveBirds()
{
    for(int i=0; i<40; i++)
    {
        if(birds[i].alive)
        {
            birds[i].x += birds[i].xSpeed;
            if(birds[i].x > ((256-32)<<8))
                birds[i].xSpeed = -abs(birds[i].xSpeed);
            if(birds[i].x < 0)
                birds[i].xSpeed = abs(birds[i].xSpeed);

            birds[i].lastbomb++;

            if(birds[i].lastbomb > 60 
                    && (abs(birds[i].x - man.x) < 20<<8) 
                    && rand()%5 == 0)
            {
                char buf[128];
                sprintf(buf, "bomb: %d, %d", (abs(birds[i].x - man.x)), (abs(birds[i].x - man.x))>>8);
                print(buf);
                dropBomb(&birds[i]);
            }

            MoveSprite(&birds[i]);
        }
    }
}
void WaitForVblank()
{
        while(REG_VCOUNT!=192);
        while(REG_VCOUNT==192);
}

void addShot(int x, int y)
{
    int n=-1;
    for(int i=0; i<40; i++)
    {
        if(shots[i].alive == false)
        {
            n=i;
            break;
        }
    }
    if(n==-1)
        return;

    shots[n].alive = true;
    shots[n].oam = n + 40;

    shots[n].attribute[0] = ATTR0_COLOR_256 | ATTR0_SQUARE;
    shots[n].attribute[1] = ATTR1_SIZE_8;
    shots[n].attribute[2] = SPR_BULLET;


    int spd = 5<<8;
    //var shot = attachMovie("shot", "shot" + nextDepth, nextDepth++);
    shots[n].x = man.x+4;
    shots[n].y = man.y+4;
    int dx = (man.x + 8) - x;
    int dy = (man.y + 8) - y;
    int dist = ((dx>>8) * (dx>>8)) + ((dy>>8) * (dy>>8));
    char buf[256];
//    sprintf(buf, "dist1 is: %d", dist);
//    print(buf);
    dist = sqrt32(dist);
//    sprintf(buf, "dist2 is: %d", dist);
//    print(buf);
    dist = dist<<8;
//    sprintf(buf, "dist3 is: %d", dist);
//    print(buf);
//    sprintf(buf, "dx: %d / %d / %d", dx, dx>>8, (dx>>8) * (dx>>8));
//    print(buf);
//    sprintf(buf, "dy: %d / %d / %d", dy, dy>>8, (dy>>8) * (dy>>8));
//    print(buf);
        
    shots[n].xSpeed = (-dx*spd) / dist;
    shots[n].ySpeed = (-dy*spd) / dist;


    MoveSprite(&shots[n]);
}

void leavePup(Sprite *bird)
{
  int n=-1;
    for(int i=0; i<5; i++)
    {
        if(pups[i].alive == false)
        {
            n=i;
            break;
        }
    }
    if(n==-1)
        return;

    pups[n].alive = true;
    pups[n].oam = n + 120;

    pups[n].attribute[0] = ATTR0_COLOR_256 | ATTR0_SQUARE;
    pups[n].attribute[1] = ATTR1_SIZE_16;

    int type = rand() % 3;

    if(type == 0)
        pups[n].attribute[2] = SPR_POWER_HP;
    if(type == 1)
        pups[n].attribute[2] = SPR_POWER_SHOT;
    if(type == 2)
        pups[n].attribute[2] = SPR_POWER_SPD;

    pups[n].x = bird->x + (8<<8);
    pups[n].y = bird->y;
    pups[n].xSpeed = 0;
    pups[n].ySpeed = 1<<8;

    MoveSprite(&pups[n]);


}

void movePups()
{
    for(int p=0; p<5; p++)
    {
        if(pups[p].alive)
        {
            if(collided(pups[p].x, pups[p].y, 16<<8, 16<<8,
                        man.x, man.y, 16<<8, 32<<8))
            {
                pups[p].alive = false;
                int t = pups[p].attribute[2];
                if(t == SPR_POWER_HP)
                    man.hp++;
                if(t == SPR_POWER_SHOT)
                    mangun++;
                if(t == SPR_POWER_SPD)
                    manspeed++;
            }
            else
            {
                pups[p].y += pups[p].ySpeed;
                pups[p].x += pups[p].xSpeed;
                if(pups[p].y > (192*2)<<8)
                    pups[p].alive = false;
            }
            MoveSprite(&pups[p]);
        }
    }
}
                
    
void moveShots()
{
    for(int i=0; i<40; i++)
    {
        if(shots[i].alive)
        {
            shots[i].x += shots[i].xSpeed;
            shots[i].y += shots[i].ySpeed;
            if(shots[i].x > ((256)<<8))
                shots[i].alive = false;
            if(shots[i].x < -8)
                shots[i].alive = false;
            if(shots[i].y < -8)
                shots[i].alive = false;
            MoveSprite(&shots[i]);
        }
    }
}

void checkHits()
{
    for(int s=0; s<40; s++)
    {
        if(shots[s].alive)
        {
        
            for(int b=0; b<40; b++)
            {
                if(birds[b].alive)
                {
                    if(collided(shots[s].x, shots[s].y, 8<<8, 8<<8,  
                                birds[b].x, birds[b].y, 32<<8, 16<<8))
                    {
                        birds[b].alive = false;
                        shots[s].alive = false;
                        if(rand()%10 == 5)
                            leavePup(&birds[b]);
                        MoveSprite(&birds[b]);
                        MoveSprite(&shots[s]);
                        score++;
                        hits++;
                    }
                }
            }
        }
    }
}

void checkBombs()
{
    for(int bm=0; bm<40; bm++)
    {
        if(bombs[bm].alive)
        {
            if(collided(bombs[bm].x, bombs[bm].y, 8<<8, 8<<8,
                        man.x, man.y, 16<<8, 32<<8))
            {
                bombs[bm].alive = false;
                MoveSprite(&bombs[bm]);
                

                man.hp--;
            }

        }
    }
}

//////////////////////////////////////////////////////////////////////

int main(int argc, char ** argv) {
  // Red main screen, blue sub-screen

  // Turn on the screens and 2D cores and switch to mode 0
    powerON(POWER_ALL);
    videoSetMode( MODE_0_2D | DISPLAY_SPR_1D_LAYOUT 
            | DISPLAY_SPR_ACTIVE | DISPLAY_BG0_ACTIVE );
    videoSetModeSub( MODE_0_2D | DISPLAY_SPR_1D_LAYOUT 
            | DISPLAY_SPR_ACTIVE | DISPLAY_BG0_ACTIVE );

    vramSetMainBanks(VRAM_A_MAIN_BG,VRAM_B_MAIN_SPRITE,VRAM_C_SUB_BG,VRAM_D_SUB_SPRITE);

	irqInitHandler(irqDefaultHandler);
	irqSet(IRQ_VBLANK, 0);

    uint16* map0 = (uint16*)SCREEN_BASE_BLOCK(0);
    uint16* map1 = (uint16*)SCREEN_BASE_BLOCK_SUB(0);

    COPY16((uint16*)font_raw, (uint16 *)CHAR_BASE_BLOCK_SUB(1), 1536);
    COPY16((uint16*)font_pal, (uint16 *)BG_PALETTE_SUB, 16);

    COPY16((uint16*)font_raw, (uint16 *)CHAR_BASE_BLOCK(1), 1536);
    COPY16((uint16*)font_pal, (uint16 *)BG_PALETTE, 16);

    BG0_CR = BG_COLOR_16| (0 << SCREEN_SHIFT) | (1 << CHAR_SHIFT);
    SUB_BG0_CR = BG_COLOR_16 | (0 << SCREEN_SHIFT) | (1 << CHAR_SHIFT);


  PALETTE[0] = RGB15(31, 31, 31);
  PALETTE[512] = RGB15(31, 31, 31);
  BG_PALETTE[0] = RGB15(31, 31, 31);
  BG_PALETTE_SUB[0] = RGB15(31, 31, 31);

	//Load palette and sprite data (main screen)
        
        aP_depack((void *)bullet_raw, (void *)SPRITE_GFX);
        aP_depack((void *)bird_raw, (void *)(SPRITE_GFX+32));
        aP_depack((void *)man_raw, (void *)(SPRITE_GFX+256+32));
        aP_depack((void *)poop_raw, (void *)(SPRITE_GFX+256+256+32));

        aP_depack((void *)power_hp_raw, (void *)(SPRITE_GFX+256+256+32+64));
        aP_depack((void *)power_shot_raw, (void *)(SPRITE_GFX+256+256+32+64+128));
        aP_depack((void *)power_spd_raw, (void *)(SPRITE_GFX+256+256+32+64+128+128));

        aP_depack((void *)sprites_pal, (void *)SPRITE_PALETTE);


	//Load palette and sprite data (sub screen)
        aP_depack((void *)bullet_raw, (void *)SPRITE_GFX_SUB);
        aP_depack((void *)bird_raw, (void *)(SPRITE_GFX_SUB+32));
        aP_depack((void *)man_raw, (void *)(SPRITE_GFX_SUB+256+32));
        aP_depack((void *)poop_raw, (void *)(SPRITE_GFX_SUB+256+256+32));

        aP_depack((void *)power_hp_raw, (void *)(SPRITE_GFX_SUB+256+256+32+64));
        aP_depack((void *)power_shot_raw, (void *)(SPRITE_GFX_SUB+256+256+32+64+128));
        aP_depack((void *)power_spd_raw, (void *)(SPRITE_GFX_SUB+256+256+32+64+128+128));
        aP_depack((void *)sprites_pal, (void *)SPRITE_PALETTE_SUB);

	//Setup man
  while(1)
    {
	initOAM();

	man.attribute[0] = ATTR0_COLOR_256 | ATTR0_TALL;
	man.attribute[1] = ATTR1_SIZE_32;
	man.attribute[2] = SPR_MAN;

	man.x = 128 << 8;
	man.y = (192+192-33) << 8;
	man.xSpeed = 0;
	man.ySpeed = 0;
	man.oam = 127;
        man.alive = true;
        man.hp = 10;
        MoveSprite(&man);

        hits = score = fired = 0;


        updateOAM();

        int n=0;
        int birdcounter=0;
        int bc=120;

          for(int i=0; i<40; i++)
        {
            birds[i].alive = bombs[i].alive = shots[i].alive = false;
        }



      while (man.hp > 0) 
      {
          if(KEY_TOUCH)
          {
              if(n == 0)
              {
                    int x = (TOUCH_X-4);
                    int y = (TOUCH_Y+192-4);
                    n = 1;
                    for(int i=0; i<mangun; i++)
                    {
                        int dx = (i - (mangun/2))*5 + 5;
//                        int dx = (20*(i - (mangun / 2))) + 10;
                        addShot((x - (dx))<<8, (y<<8));
//                        addShot(_xmouse - (dx * 20), _ymouse);
                    }
                    fired++;
              }
              rand();
                
          }
          else 
              n = 0;

          birdcounter++;
          if(birdcounter > bc)
          {
              birdcounter=0;
              bc--;
              if(bc < 20)
                  bc = 20;
              addBird();
          }

    //      if(READ_KEYS & KEY_A)
      //        addBird();

          if(READ_KEYS & KEY_RIGHT)
          {
              man.x +=manspeed << 8;
              MoveSprite(&man);
          }
          if(READ_KEYS & KEY_LEFT)
          {
              man.x -=manspeed << 8;
              MoveSprite(&man);
          }
          if(READ_KEYS & KEY_UP)
          {
              man.y -=1<< 8;
              MoveSprite(&man);
          }
          if(READ_KEYS & KEY_DOWN)
          {
              man.y +=1<< 8;
              MoveSprite(&man);
          }

          moveBirds();
          moveShots();
          moveBombs();
          movePups();
          checkHits();
          checkBombs();

          writeText(0,0, "HP: ", map1);
          writeNum(3,0, man.hp, map1);
          writeText(10, 0, "SCORE: ", map1);
          writeText(22, 0, "ACC: ", map1);
          if(fired > 0)
          {
              writeNum(17, 0, (score * 50 * hits)/fired, map1);
              writeNum(26, 0, (hits * 100)/fired, map1);
          }
          else
          {
              writeNum(17, 0, 0, map1);
              writeNum(26, 0, 0, map1);
          }


          WaitForVblank();
        updateOAM();
      }

      initOAM();
      updateOAM();
      writeText(10, 10, "GAME OVER", map1);
      writeText(3, 10, "PRESS START TO PLAY AGAIN", map0);

      while(READ_KEYS & KEY_START) ;
      while(!(READ_KEYS & KEY_START)) ;

      writeText(10, 10, "         ", map1);
      writeText(3, 10, "                         ", map0);

    }
  return 0;
}

//////////////////////////////////////////////////////////////////////
